@@ -1,17 +1,19 @@
 """
 This module provides IO primitives for working with cross-API RedVox data.
 """
+import enum
 from collections import defaultdict
 from dataclasses import dataclass, field
 from datetime import datetime, timedelta
 from glob import glob
+import numpy as np
+import json
 import os.path
 import multiprocessing
 import multiprocessing.pool
+import tempfile
 from pathlib import Path, PurePath
-import pickle
-import json
-from shutil import copy2, move
+from shutil import copy2, move, rmtree
 from typing import (
     Any,
     Dict,
@@ -27,6 +29,8 @@ from typing import (
 import lz4.frame
 
 from redvox.api900.reader import read_rdvxz_file, read_buffer
+from redvox.api900.reader_utils import calculate_uncompressed_size
+from redvox.common import api_conversions as ac
 from redvox.api1000.common.common import check_type
 from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM
 from redvox.api1000.proto.redvox_api_m_pb2 import RedvoxPacketM
@@ -42,10 +46,295 @@ from redvox.common.parallel_utils import maybe_parallel_map
 
 if TYPE_CHECKING:
     from redvox.api900.wrapped_redvox_packet import WrappedRedvoxPacket
-    from redvox.common.data_window import DataWindow
     from redvox.api900.lib.api900_pb2 import RedvoxPacket
 
 
+def remove_dir_contents(dir_path: Path):
+    """
+    removes all contents of the directory specified by dir_path
+
+    :param dir_path: path to directory to remove files from
+    """
+    if dir_path.is_dir():
+        for entry in os.listdir(dir_path):
+            rmv_path = os.path.join(dir_path, entry)
+            if os.path.isdir(rmv_path):
+                rmtree(rmv_path)
+            else:
+                os.remove(rmv_path)
+    else:
+        print(f"{dir_path} is not a directory; cannot remove contents!")
+
+
+class FileSystemSaveMode(enum.Enum):
+    """
+    Enumeration of saving methodology
+    """
+    MEM = 0   # save using memory
+    TEMP = 1  # save using temporary directory
+    DISK = 2  # save using path on disk
+
+    @staticmethod
+    def get_save_mode(use_temp: bool, use_disk: bool) -> "FileSystemSaveMode":
+        """
+        :param use_temp: use temporary directory
+        :param use_disk: use path on disk
+        :return: the mode used to save (use_temp is evaluated before use_disk)
+        """
+        if use_temp:
+            return FileSystemSaveMode.TEMP  # use_temp takes priority
+        elif use_disk:
+            return FileSystemSaveMode.DISK  # if here, use_temp is always false
+        return FileSystemSaveMode.MEM
+
+
+class FileSystemWriter:
+    """
+    This class holds basic information about writing and reading objects from a file system
+    If user does not enable saving to disk, we use a temporary directory to store large files
+
+    Properties:
+        file_name: str, the name of the file (do not include extension)
+
+        file_ext: str, the extension used by the file (do not include the .)  Default "NONE"
+
+        base_dir: str, the directory to save the file to.  Default "." (current dir)
+
+    Protected:
+        _save_mode: FileSystemSaveMode, determines how files get saved
+
+        _temp_dir: TemporaryDirectory, temporary directory for large files when not saving to disk
+    """
+
+    def __init__(self, file_name: str, file_ext: str = "none", base_dir: str = ".",
+                 save_mode: FileSystemSaveMode = FileSystemSaveMode.MEM):
+        """
+        initialize FileSystemWriter
+
+        :param file_name: name of file
+        :param file_ext: extension of file, default "none"
+        :param base_dir: directory to save file to, default "." (current dir)
+        :param save_mode: determines how to save files to system, default MEM (no save, use RAM)
+        """
+        self.file_name: str = file_name
+        self.file_extension: str = file_ext.lower()
+        self.base_dir: str = base_dir
+        self._save_mode: FileSystemSaveMode = save_mode
+        self._temp_dir = tempfile.TemporaryDirectory()
+
+    def __repr__(self):
+        return f"file_name: {self.file_name}, " \
+               f"extension: {self.file_extension}, " \
+               f"base_dir: {self.base_dir}, " \
+               f"save_mode: {self._save_mode.value}"
+
+    def __str__(self):
+        return f"file_name: {self.file_name}, " \
+               f"extension: {self.file_extension}, " \
+               f"base_dir: {self.base_dir}, " \
+               f"save_mode: {self._save_mode.name}"
+
+    def is_use_temp(self) -> bool:
+        """
+        :return: if writing to temp dir
+        """
+        return self._save_mode == FileSystemSaveMode.TEMP
+
+    def set_use_temp(self, use_temp: bool):
+        """
+        :param use_temp: if true, sets mode to use temp dir, otherwise no change
+        """
+        if use_temp:
+            self._save_mode = FileSystemSaveMode.TEMP
+
+    def get_temp(self) -> str:
+        """
+        :return: path of temp directory
+        """
+        return self._temp_dir.name
+
+    def is_use_disk(self) -> bool:
+        """
+        :return: if writing to path on disk
+        """
+        return self._save_mode == FileSystemSaveMode.DISK
+
+    def set_use_disk(self, use_disk: bool):
+        """
+        :param use_disk: if true, sets mode to use the disk, otherwise no change
+        """
+        if use_disk:
+            self._save_mode = FileSystemSaveMode.DISK
+
+    def is_use_mem(self) -> bool:
+        """
+        :return: if writing data to memory
+        """
+        return self._save_mode == FileSystemSaveMode.MEM
+
+    def set_use_mem(self, use_mem: bool):
+        """
+        :param use_mem: if true, sets mode to use the system's RAM, otherwise no change
+        """
+        if use_mem:
+            self._save_mode = FileSystemSaveMode.MEM
+
+    def is_save_disk(self) -> bool:
+        """
+        :return: if writing data to disk (temp dir or user defined path) instead of using memory
+        """
+        return self._save_mode != FileSystemSaveMode.MEM
+
+    def save_dir(self) -> str:
+        """
+        :return: directory where file would be saved based on save mode; returns empty string if saving to memory
+        """
+        if self.is_use_disk():
+            return self.base_dir
+        elif self.is_use_temp():
+            return self._temp_dir.name
+        return ""
+
+    def set_save_mode(self, save_mode: FileSystemSaveMode):
+        """
+        set the save mode
+
+        :param save_mode: updated save mode
+        """
+        self._save_mode = save_mode
+
+    def save_mode(self) -> FileSystemSaveMode:
+        """
+        :return: the save mode
+        """
+        return self._save_mode
+
+    def full_name(self) -> str:
+        """
+        :return: file name with extension
+        """
+        return f"{self.file_name}.{self.file_extension}"
+
+    def full_path(self) -> str:
+        """
+        :return: the full path to where the file would be written
+        """
+        return os.path.join(self.save_dir(), self.full_name())
+
+    def set_name_and_extension(self, name: str, ext: str):
+        """
+        set the name and extension of the output file.  Do not include the . for the extension
+        :param name: file name
+        :param ext: file extension
+        """
+        self.file_name = name
+        self.file_extension = ext
+
+    def set_name(self, name: str):
+        """
+        set the name of the output file.
+        :param name: file name
+        """
+        self.file_name = name
+
+    def set_extension(self, ext: str):
+        """
+        set the extension of the output file.  Do not include the . for the extension
+        :param ext: file extension
+        """
+        self.file_extension = ext
+
+    def json_file_name(self) -> str:
+        """
+        :return: file name with .json extension
+        """
+        return f"{self.file_name}.json"
+
+    def json_path(self) -> Path:
+        """
+        :return: full path to json file
+        """
+        return Path(self.save_dir()).joinpath(self.json_file_name())
+
+    def create_dir(self):
+        """
+        if saving to disk, remove the directory if it exists,
+        then create an empty directory to save things into
+        if saving to temp dir, remove any files in the temp dir
+        """
+        if self.is_use_disk():
+            if os.path.exists(self.save_dir()):
+                remove_dir_contents(Path(self.save_dir()))
+            else:
+                os.makedirs(self.save_dir())
+        elif self.is_use_temp():
+            remove_dir_contents(Path(self._temp_dir.name))
+
+    def __del__(self):
+        """
+        remove temp dir
+        """
+        self._temp_dir.cleanup()
+
+    def as_dict(self) -> dict:
+        """
+        :return: FileSystemWriter as dictionary
+        """
+        return {
+            "file_name": self.file_name,
+            "file_extension": self.file_extension,
+            "base_dir": self.base_dir,
+            "save_mode": self._save_mode.name
+        }
+
+    @staticmethod
+    def from_dict(data_dict: Dict) -> "FileSystemWriter":
+        """
+        :param data_dict: dictionary to convert to FileSystemWriter
+        :return: FileSystemWriter from dict
+        """
+        return FileSystemWriter(data_dict["file_name"], data_dict["file_extension"], data_dict["base_dir"],
+                                FileSystemSaveMode[data_dict["save_mode"]])
+
+
+def dict_to_json(dct: dict) -> str:
+    """
+    :param dct: dictionary to convert to json
+    :return: dictionary as json string
+    """
+    return json.dumps(dct)
+
+
+def json_to_dict(json_str: str) -> Dict:
+    """
+    :param json_str: string of json to convert to dictionary
+    :return: json string as a dictionary
+    """
+    return json.loads(json_str)
+
+
+def json_file_to_dict(file_path: str) -> Dict:
+    """
+    :param file_path: full path of file to load data from.
+    :return: json file as python dictionary
+    """
+    with open(file_path, "r") as f_p:
+        return json_to_dict(f_p.read())
+
+
+def get_json_file(file_dir: str) -> Optional[str]:
+    """
+    Finds the first json file in the file_dir specified or None if there is no file
+
+    :param file_dir: directory to find json file in
+    :return: full name of first json file in the directory or None if no files found
+    """
+    file_names = glob(os.path.join(file_dir, "*.json"))
+    if len(file_names) < 1:
+        return None
+    return Path(file_names[0]).name
+
+
 def _is_int(value: str) -> Optional[int]:
     """
     Tests if a given str is a valid integer. If it is, the integer is returned, if it is not, None is returned.
@@ -81,6 +370,8 @@ class IndexEntry:
     date_time: datetime
     extension: str
     api_version: ApiVersion
+    compressed_file_size_bytes: int = 0
+    decompressed_file_size_bytes: int = 0
 
     @staticmethod
     def from_path(path_str: str, strict: bool = True) -> Optional["IndexEntry"]:
@@ -132,7 +423,7 @@ class IndexEntry:
                 return None
             full_path = path_str
 
-        return IndexEntry(full_path, station_id, date_time, ext, api_version)
+        return IndexEntry(full_path, station_id, date_time, ext, api_version)._set_compressed_decompressed_lz4_size()
 
     @staticmethod
     def from_native(entry) -> "IndexEntry":
@@ -147,8 +438,8 @@ class IndexEntry:
             entry.station_id,
             dt_us(entry.date_time),
             entry.extension,
-            ApiVersion.from_str(entry.api_version),
-        )
+            ApiVersion.from_str(entry.api_version)
+        )._set_compressed_decompressed_lz4_size()
 
     def to_native(self):
         import redvox_native
@@ -158,10 +449,27 @@ class IndexEntry:
             self.station_id,
             us_dt(self.date_time),
             self.extension,
-            self.api_version.value,
+            self.api_version.value
         )
         return entry
 
+    def _set_compressed_decompressed_lz4_size(self):
+        """
+        set the compressed and decompressed file size in bytes of an lz4 file being read by the IndexEntry.
+        default is 0 for both file sizes
+
+        :return: updated self
+        """
+        if os.path.exists(self.full_path):
+            self.compressed_file_size_bytes = os.path.getsize(self.full_path)
+            with open(self.full_path, "rb") as fp:
+                if self.api_version == ApiVersion.API_1000:
+                    header = lz4.frame.get_frame_info(fp.read())
+                    self.decompressed_file_size_bytes = header["content_size"]
+                elif self.api_version == ApiVersion.API_900:
+                    self.decompressed_file_size_bytes = calculate_uncompressed_size(fp.read())
+        return self
+
     def read(self) -> Optional[Union[WrappedRedvoxPacketM, "WrappedRedvoxPacket"]]:
         """
         Reads, decompresses, deserializes, and wraps the RedVox file pointed to by this entry.
@@ -415,6 +723,7 @@ class IndexStationSummary:
     total_packets: int
     first_packet: datetime
     last_packet: datetime
+    single_packet_decompressed_size_bytes: int
 
     @staticmethod
     def from_entry(entry: IndexEntry) -> "IndexStationSummary":
@@ -430,6 +739,7 @@ class IndexStationSummary:
             1,
             first_packet=entry.date_time,
             last_packet=entry.date_time,
+            single_packet_decompressed_size_bytes=entry.decompressed_file_size_bytes
         )
 
     def update(self, entry: IndexEntry) -> None:
@@ -564,7 +874,7 @@ class Index:
         entries: List[IndexEntry] = list(
             map(IndexEntry.from_native, index_native.entries)
         )
-        return Index(entries)
+        return Index(entries)._set_decompressed_file_size()
 
     def to_native(self):
         import redvox_native
@@ -573,6 +883,35 @@ class Index:
         native_index.entries = list(map(IndexEntry.to_native, self.entries))
         return native_index
 
+    def max_decompressed_file_size(self) -> int:
+        """
+        :return: the maximum decompressed file size in the entries
+        """
+        return max([fi.decompressed_file_size_bytes for fi in self.entries]) if len(self.entries) > 0 else np.nan
+
+    def get_decompressed_file_size(self) -> int:
+        """
+        :return: the decompressed size of the first file in the list of entries
+        """
+        if len(self.entries) == 0:
+            return np.nan
+        if self.entries[0].decompressed_file_size_bytes == 0 and os.path.exists(self.entries[0].full_path):
+            with lz4.frame.open(self.entries[0].full_path, 'rb') as fr:
+                return len(fr.read())
+        return self.entries[0].decompressed_file_size_bytes
+
+    def _set_decompressed_file_size(self) -> "Index":
+        """
+        updates the decompressed size of all entries if the maximum decompressed size is 0, otherwise makes no changes
+
+        :return: updated self
+        """
+        if self.max_decompressed_file_size() == 0:
+            new_size = self.get_decompressed_file_size()
+            for ie in self.entries:
+                ie.decompressed_file_size_bytes = new_size
+        return self
+
     def sort(self) -> None:
         """
         Sorts the entries stored in this index.
@@ -589,6 +928,7 @@ class Index:
         :param entries: Entries to append.
         """
         self.entries.extend(entries)
+        self._set_decompressed_file_size()
 
     def summarize(self) -> IndexSummary:
         """
@@ -651,6 +991,66 @@ class Index:
         """
         return list(self.stream(read_filter))
 
+    def files_size(self) -> float:
+        """
+        :return: sum of file size in bytes of index
+        """
+        return float(np.sum([entry.decompressed_file_size_bytes for entry in self.entries]))
+
+    def read_contents(self) -> List[RedvoxPacketM]:
+        """
+        read all the files in the index
+
+        :return: list of RedvoxPacketM, converted from API 900 if necessary
+        """
+        result: List[RedvoxPacketM] = []
+
+        # Iterate over the API 900 packets in a memory efficient way
+        # and convert to API 1000
+        # noinspection PyTypeChecker
+        for packet_900 in self.stream_raw(
+                ReadFilter.empty().with_api_versions({ApiVersion.API_900})
+        ):
+            # noinspection Mypy
+            result.append(
+                ac.convert_api_900_to_1000_raw(packet_900)
+            )
+
+        # Grab the API 1000 packets
+        # noinspection PyTypeChecker
+        for packet in self.stream_raw(
+                ReadFilter.empty().with_api_versions({ApiVersion.API_1000})
+        ):
+            # noinspection Mypy
+            result.append(packet)
+
+        return result
+
+    def read_first_packet(self) -> Optional[RedvoxPacketM]:
+        """
+        read the first packet of the index
+
+        :return: single RedvoxPacketM, converted from API 900 if necessary or None if no packet to read
+        """
+        # Grab the API 1000 packets
+        # noinspection PyTypeChecker
+        for packet in self.stream_raw(
+                ReadFilter.empty().with_api_versions({ApiVersion.API_1000})
+        ):
+            # noinspection Mypy
+            return packet
+
+        # Iterate over the API 900 packets in a memory efficient way
+        # and convert to API 1000
+        # noinspection PyTypeChecker
+        for packet_900 in self.stream_raw(
+                ReadFilter.empty().with_api_versions({ApiVersion.API_900})
+        ):
+            # noinspection Mypy
+            return ac.convert_api_900_to_1000_raw(packet_900)
+
+        return None
+
 
 # The following constants are used for identifying valid RedVox API 900 and API 1000 structured directory layouts.
 __VALID_YEARS: Set[str] = {f"{i:04}" for i in range(2015, 2031)}
@@ -1219,162 +1619,3 @@ def sort_unstructured_redvox_data(
             move(value.full_path, file_out_dir)
 
     return True
-
-
-def json_to_data_window(json_str: str) -> Dict:
-    """
-    load a data window from json string
-
-    :param json_str: json string to read
-    :return: a dictionary of a data window
-    """
-    return json.loads(json_str)
-
-
-@dataclass
-class DataWindowSerializationResult:
-    path: str
-    serialized_bytes: int
-    compressed_bytes: int
-
-
-def serialize_data_window(
-        data_window: "DataWindow",
-        base_dir: str = ".",
-        file_name: Optional[str] = None,
-        compression_factor: int = 4,
-) -> Path:
-    """
-    Serializes and compresses a DataWindow to a file.
-
-    :param data_window: The data window to serialize and compress.
-    :param base_dir: The base directory to write the serialized file to (default=.).
-    :param file_name: The optional file name. If None, a default filename with the following format is used:
-                      [start_ts]_[end_ts]_[num_stations].pkl.lz4
-    :param compression_factor: A value between 1 and 12. Higher values provide better compression, but take longer.
-                               (default=4).
-    :return: The path to the written file.
-    """
-
-    _file_name: str = (
-        file_name
-        if file_name is not None
-        else f"{int(data_window.start_datetime.timestamp())}"
-             f"_{int(data_window.end_datetime.timestamp())}"
-             f"_{len(data_window.station_ids)}.pkl.lz4"
-    )
-
-    file_path: Path = Path(base_dir).joinpath(_file_name)
-
-    with lz4.frame.open(
-            file_path, "wb", compression_level=compression_factor
-    ) as compressed_out:
-        pickle.dump(data_window, compressed_out)
-        compressed_out.flush()
-        return file_path.resolve(False)
-
-
-def deserialize_data_window(path: str) -> "DataWindow":
-    """
-    Decompresses and deserializes a DataWindow written to disk.
-
-    :param path: Path to the serialized and compressed data window.
-    :return: An instance of a DataWindow.
-    """
-    with lz4.frame.open(path, "rb") as compressed_in:
-        return pickle.load(compressed_in)
-
-
-def json_file_to_data_window(base_dir: str, file_name: str) -> Dict:
-    """
-    load a data window from json written to disk
-
-    :param base_dir: directory where json file is saved
-    :param file_name: name of json file to load
-    :return: a dictionary representing a json-ified data window
-    """
-    with open(Path(base_dir).joinpath(file_name), "r") as r_f:
-        return json_to_data_window(r_f.read())
-
-
-def data_window_to_json(
-    data_win: "DataWindow",
-    base_dir: str = ".",
-    file_name: Optional[str] = None,
-    compression_format: str = "lz4",
-) -> str:
-    """
-    Converts a data window to json format
-
-    :param data_win: the data window object to convert
-    :param base_dir: the base directory to write the data file to
-    :param file_name: the data object's optional base file name.  Do not include a file extension.
-                        If None, a default file name is created using this format:
-                        [start_ts]_[end_ts]_[num_stations].[compression_type]
-    :param compression_format: the compression format to use.  default lz4
-    :return: The path to the written file
-    """
-    _file_name: str = (
-        file_name
-        if file_name is not None
-        else f"{int(data_win.start_datetime.timestamp())}"
-        f"_{int(data_win.end_datetime.timestamp())}"
-        f"_{len(data_win.station_ids)}"
-    )
-    base_dir = os.path.join(base_dir, "dw")
-    if not os.path.exists(base_dir):
-        os.makedirs(base_dir)
-    if compression_format == "lz4":
-        dfp = str(
-            serialize_data_window(
-                data_win, base_dir, _file_name + ".pkl.lz4"
-            ).resolve()
-        )
-    else:
-        dfp = os.path.join(base_dir, _file_name + ".pkl")
-        with open(dfp, "wb") as compressed_out:
-            pickle.dump(data_win, compressed_out)
-    data_win_dict = {
-        "start_datetime": us_dt(data_win.start_datetime)
-        if data_win.start_datetime
-        else None,
-        "end_datetime": us_dt(data_win.end_datetime) if data_win.end_datetime else None,
-        "station_ids": list(data_win.station_ids),
-        "compression_format": compression_format,
-        "file_name": _file_name,
-    }
-    return json.dumps(data_win_dict)
-
-
-def data_window_to_json_file(
-    data_window: "DataWindow",
-    base_dir: str = ".",
-    file_name: Optional[str] = None,
-    compression_format: str = "lz4",
-) -> Path:
-    """
-    Converts a data window to json format.
-
-    :param data_window: the data window object to convert
-    :param base_dir: the base directory to write the json file to
-    :param file_name: the optional base file name.  Do not include a file extension.
-                        If None, a default file name is created using this format:
-                        [start_ts]_[end_ts]_[num_stations].json
-    :param compression_format: the type of compression to use.  default lz4
-    :return: The path to the written file
-    """
-    _file_name: str = (
-        file_name
-        if file_name is not None
-        else f"{int(data_window.start_datetime.timestamp())}"
-        f"_{int(data_window.end_datetime.timestamp())}"
-        f"_{len(data_window.station_ids)}"
-    )
-    file_path: Path = Path(base_dir).joinpath(f"{_file_name}.json")
-    with open(file_path, "w") as f_p:
-        f_p.write(
-            data_window_to_json(
-                data_window, base_dir, file_name, compression_format
-            )
-        )
-        return file_path.resolve(False)
